from oarepo_model_builder.model_preprocessors import ModelPreprocessor


class InvenioBaseClassesModelPreprocessor(ModelPreprocessor):
    TYPE = "invenio_base_classes"

    def transform(self, schema, settings):
        self.set_default_and_append_if_not_present(
            schema.current_model,
            "record-resource-bases",
            [],
            schema.current_model.get(
                "record-resource-parent-class",
                "invenio_records_resources.resources.RecordResource",
            ),
        )
        self.set_default_and_append_if_not_present(
            schema.current_model,
            "record-resource-config-bases",
            [],
            schema.current_model.get(
                "record-resource-config-parent-class",
                "invenio_records_resources.resources.RecordResourceConfig",
            ),
        )
        self.set_default_and_append_if_not_present(
            schema.current_model,
            "record-service-bases",
            [],
            schema.current_model.get(
                "record-service-parent-class",
                "invenio_records_resources.services.RecordService",
            ),
        )
        self.set_default_and_append_if_not_present(
            schema.current_model,
            "record-service-config-bases",
            [],
            schema.current_model.get(
                "record-service-config-parent-class",
                "invenio_records_resources.services.RecordServiceConfig",
            ),
        )
        self.set_default_and_append_if_not_present(
            schema.current_model,
            "record-bases",
            [],
            schema.current_model.get(
                "record-parent-class", "invenio_records_resources.records.api.Record"
            ),
        )
        self.set_default_and_append_if_not_present(
            schema.current_model,
            "record-metadata-bases",
            [],
            schema.current_model.get(
                "record-metadata-parent-class",
                "invenio_records.models.RecordMetadataBase",
            ),
        )



import pytest


@pytest.fixture(autouse=True)
def clear_datatypes():
    from oarepo_model_builder.datatypes import datatypes

    datatypes.clear_cache()


@pytest.fixture(autouse=True)
def clear_validation():
    from oarepo_model_builder.validation import model_validator

    model_validator.clear_cache()


def pytest_configure():
    pass
    # logging.basicConfig(level=logging.ERROR)
    # logger = logging.getLogger("oarepo_model_builder.cst")
    # logger.setLevel(logging.DEBUG)



import enum

from marshmallow import fields

from .utils import ExtendablePartSchema


class OrderEnum(enum.Enum):
    asc = "asc"
    desc = "desc"


class PropertySortable(ExtendablePartSchema):
    key = fields.String(required=False)
    order = fields.Enum(
        OrderEnum,
        dump_default=OrderEnum.asc,
    )



from . import process
from .json_base import JSONBaseBuilder


class ExtendBuilder(JSONBaseBuilder):
    TYPE = "extend"
    output_file_type = "json"

    @process("**")
    def model_element(self):
        self.model_element_enter()
        self.build_children()
        self.model_element_leave()

    def enter_model(self):
        # build children are handled in model_element, so no default handling here
        pass

    def output_primitive(self, top, data):
        if data is not None and not isinstance(data, (str, float, int, bool)):
            data = str(data)
        return super().output_primitive(top, data)

    def begin(self, schema, settings):
        super(JSONBaseBuilder, self).begin(schema, settings)
        self.output = self.builder.get_output(self.output_file_type, "model.json5")
        self.output.primitive("type", "object")

    def finish(self):
        # force clean output
        super().finish()
        self.output.force_clean_output()



import re
from typing import List

from marshmallow import fields
from marshmallow.exceptions import ValidationError

from .datatypes import DataType, Import


def validate_regex(value):
    if not value:
        return
    try:
        re.compile(value)
    except Exception as e:
        raise ValidationError(
            f"Regex '{value}' is not valid. Reported error: '{str(e)}'"
        )


class StringDataType(DataType):
    schema_type = "string"
    marshmallow_field = "ma_fields.String"

    class ModelSchema(DataType.ModelSchema):
        minLength = fields.Integer(required=False)
        maxLength = fields.Integer(required=False)
        regex = fields.String(required=False, validate=validate_regex)
        enum = fields.List(fields.String(), required=False)

    def marshmallow_validators(self):
        validators = []
        ranges = {}
        for param, schema in (
            ("min", "minLength"),
            ("max", "maxLength"),
        ):
            if schema in self.definition:
                ranges[param] = self.definition[schema]

        if ranges:
            params = ", ".join(f"{k}={v}" for k, v in ranges.items())
            validators.append(f"ma_validate.Length({params})")

        if "pattern" in self.definition:
            pattern = self.definition["pattern"]
            pattern = pattern.replace("\\", "\\\\")
            pattern = pattern.replace('"', '\\"')
            validators.append(f'ma_validate.Regexp("{pattern}")')

        return validators

    @property
    def ui_marshmallow_field(self):
        if "enum" in self.definition:
            return "l10n.LocalizedEnum"
        return self.marshmallow_field

    def ui_marshmallow(self, **extras):
        ret = super().ui_marshmallow(**extras)
        if "enum" in self.definition:
            # TODO: allow for custom prefix
            gettext_prefix = self.model.get("package")
            ret.setdefault("arguments", []).extend(
                [
                    f'value_prefix="{gettext_prefix}"',
                ]
            )
        return ret

    def imports(self, *extra) -> List[Import]:
        if "enum" in self.definition:
            return super().imports(
                *extra,
                Import(
                    import_path="oarepo_runtime.ui.marshmallow",
                    alias="l10n",
                ),
            )
        else:
            return super().imports(*extra)


class FulltextDataType(StringDataType):
    mapping_type = "text"
    model_type = "fulltext"

    def get_facet(self, stack, parent_path, path_suffix=None):
        "fulltext can not have facet"
        return []


class KeywordDataType(StringDataType):
    mapping_type = "keyword"
    model_type = "keyword"

    @property
    def facet_class(self):
        if self.definition.get("enum"):
            return "EnumTermsFacet"
        return super().facet_class

    @property
    def facet_imports(self):
        if self.definition.get("enum"):
            return [{"import": "oarepo_runtime.facets.enum.EnumTermsFacet"}]
        return super().facet_imports


class FulltextKeywordDataType(StringDataType):
    mapping_type = "text"
    model_type = "fulltext+keyword"

    def mapping(self):
        ret = super().mapping()
        mapping_el = ret.setdefault("mapping", {})
        mapping_el.setdefault("fields", {}).setdefault("keyword", {"type": "keyword"})
        return ret

    def get_facet(self, stack, parent_path, path_suffix=None):
        return super().get_facet(stack, parent_path, (path_suffix or "") + ".keyword")


class URLDataType(StringDataType):
    mapping_type = "keyword"
    model_type = "url"



from collections import defaultdict
from functools import cached_property

import importlib_metadata
import marshmallow as ma
from marshmallow.exceptions import ValidationError

from oarepo_model_builder.validation.utils import PermissiveSchema

PROPERTY_BY_TYPE_PREFIX = "property-by-type-"
ARRAY_ITEM_BY_TYPE_PREFIX = "array-item-by-type-"


class ModelValidator:
    @cached_property
    def validator_map(self):
        ret = {}
        for ep in importlib_metadata.entry_points().select(
            group="oarepo_model_builder.validation"
        ):
            validators = ep.load()
            for validator_type, validator in validators.items():
                ret.setdefault(
                    validator_type, []
                )  # create always, regardless if the validator is empty or not - to register empty extension points
                if isinstance(validator, (list, tuple)):
                    ret[validator_type].extend(validator)
                else:
                    ret[validator_type].append(validator)
        return ret

    def clear_cache(self):
        if "validator_map" in self.__dict__:
            del self.__dict__["validator_map"]

    def validator_class(self, section="root", strict=True):
        if section.startswith(PROPERTY_BY_TYPE_PREFIX):
            validators = self.get_property_validator_class(section)
        elif section.startswith(ARRAY_ITEM_BY_TYPE_PREFIX):
            validators = self.get_array_item_validator_class(section)
        else:
            validators = tuple(
                self.validator_map[section],
            )

        if not validators:
            return PermissiveSchema

        # join fields
        if len(validators) > 1:
            merged_schema = self._merge_schemas(validators, strict)
        else:
            merged_schema = validators[0]

        return merged_schema

    def _merge_schemas(self, schemas, strict=False):
        fields = defaultdict(list)
        for cls in schemas:
            for fld_key, fld in cls._declared_fields.items():
                fields[fld_key].append(fld)

        redefined_fields = {}
        for fld_key, flds in fields.items():
            if len(flds) > 1:
                redefined = self._merge_fields(flds)
                if redefined:
                    redefined_fields[fld_key] = redefined

        # if any of the validators is extremely permissive, do not add strict meta
        parent_meta_classes = set()
        for v in schemas:
            for vv in v.mro():
                meta_class = getattr(vv, "Meta", None)
                if meta_class and meta_class not in parent_meta_classes:
                    parent_meta_classes.add(meta_class)
                    if getattr(meta_class, "unknown", None) == ma.INCLUDE:
                        strict = False
        meta_options = {}
        if strict:
            meta_options["unknown"] = ma.RAISE

        parent_meta_classes = self._clear_mro(parent_meta_classes)
        meta_class = type("Meta", tuple(parent_meta_classes), meta_options)

        # remove items that are in mro of other items
        schemas = self._clear_mro(schemas)
        return type(
            f"{'_'.join(type(x).__name__ for x in schemas)}Validator",
            tuple(schemas),
            {"Meta": meta_class, **redefined_fields},
        )

    def _clear_mro(self, types):
        return [
            s
            for s in types
            if not any(other != s and s in other.mro() for other in types)
        ]

    def _merge_fields(self, flds):
        if isinstance(flds[0], ma.fields.Nested):
            schemas = [type(x.schema) for x in flds]
            combined_schema = self._merge_schemas(schemas)
            return ma.fields.Nested(
                combined_schema,
                data_key=flds[0].data_key,
                many=flds[0].many,
                attribute=flds[0].attribute,
                allow_none=flds[0].allow_none,
            )
        else:
            return None

    def get_property_validator_class(self, section):
        datatype_name = section[len(PROPERTY_BY_TYPE_PREFIX) :]
        return self._get_datatype_validator_class(
            section, datatype_name, "object-property"
        )

    def get_array_item_validator_class(self, section):
        datatype_name = section[len(ARRAY_ITEM_BY_TYPE_PREFIX) :]
        return self._get_datatype_validator_class(section, datatype_name, "array-item")

    def _get_datatype_validator_class(
        self, section, datatype_name, extra_validation_key
    ):
        from oarepo_model_builder.datatypes import datatypes

        validators = self.validator_map.get(section, ())

        datatype_class = datatypes.get_datatype_class(datatype_name)
        if not datatype_class:
            raise ValidationError(
                f"No datatype class registered for datatype {datatype_name}"
            )
        validators = tuple(
            [
                *validators,
                *self.validator_map["property"],
                *self.validator_map[extra_validation_key],
                datatype_class.ModelSchema,
            ]
        )
        return validators

    def validate(self, data):
        validator_class = self.validator_class()
        validator = validator_class()
        return validator.dump(validator.load(data))


model_validator = ModelValidator()



from typing import Optional

from faker.providers import BaseProvider


class Provider(BaseProvider):
    def sentence(self):
        return "test"

    def random_int(self, min: int = 0, max: int = 9999, step: int = 1) -> int:
        return 1

    def random_float(self, digits: Optional[int] = None, fix_len: bool = False) -> int:
        return 1.2

    def date(self):
        return "2022-01-02"



import marshmallow as ma

from oarepo_model_builder.builders import ReplaceElement
from oarepo_model_builder.builders.jsonschema import JSONSchemaBuilder
from oarepo_model_builder.builders.mapping import MappingBuilder
from oarepo_model_builder.datatypes import DataType
from oarepo_model_builder.invenio.invenio_record_schema import (
    InvenioRecordSchemaBuilder,
)
from oarepo_model_builder.property_preprocessors import PropertyPreprocessor, process
from oarepo_model_builder.utils.deepmerge import deepmerge


class MultilangPreprocessor(PropertyPreprocessor):
    @process(
        model_builder=JSONSchemaBuilder,
        path="/properties/*",
        condition=lambda current, stack: current.type == "multilingual",
    )
    def modify_multilang_schema(self, data, stack, **kwargs):
        data["type"] = "object"
        data["properties"] = {"lang": {"type": "string"}, "value": {"type": "string"}}
        return data

    @process(
        model_builder=MappingBuilder,
        path="/properties/*",
        condition=lambda current, stack: current.type == "multilingual",
    )
    def modify_multilang_mapping(self, data, stack, **kwargs):
        raise ReplaceElement(
            {
                stack.top.key: {
                    "type": "object",
                    "properties": {
                        "lang": {"type": "keyword"},
                        "value": {"type": "text"},
                    },
                },
                stack.top.key + "_cs": {"type": "text"},
            }
        )

    @process(
        model_builder=InvenioRecordSchemaBuilder,
        path="/properties/*",
        condition=lambda current, stack: current.type == "multilingual",
    )
    def modify_multilang_marshmallow(self, data, stack, **kwargs):
        data["type"] = "object"
        deepmerge(
            data.setdefault("marshmallow", {}),
            {
                "imports": [{"import": "tests.multilang", "alias": "multilingual"}],
                "class": "multilingual.MultilingualSchema",
                "nested": True,
            },
        )
        return data


class UIValidator(ma.Schema):
    class Meta:
        unknown = ma.INCLUDE


class MultilingualDataType(DataType):
    model_type = "multilingual"
    schema_type = "object"
    mapping_type = "object"



from collections import namedtuple

from oarepo_model_builder.stack.stack import ModelBuilderStack


class ModelPreprocessor:
    def __init__(self, builder: "oarepo_model_builder.builder.ModelBuilder"):
        self.builder = builder

    def transform(self, schema, settings):
        pass

    def set(self, settings, name, callable=None):
        if callable:
            if not settings.get(name):
                settings[name] = callable()
            return

        def w(func):
            if not settings.get(name):
                settings[name] = func()

        return w

    def set_default_and_append_if_not_present(self, container, key, value, appended):
        container.setdefault(key, value)
        arr = getattr(container, key)
        if appended not in arr:
            arr.append(appended)

    def set_default_and_prepend_if_not_present(self, container, key, value, prepended):
        container.setdefault(key, value)
        arr = getattr(container, key)
        if prepended not in arr:
            arr.insert(0, prepended)


TSE = namedtuple("TransformStackEntry", "before_transform, key, data")


class DeepTransformationModelPreprocessor(ModelPreprocessor):
    def transform(self, schema, settings):
        self.begin(schema, settings)
        self.deep_transform()
        self.end()

    def begin(self, schema, settings):
        self.schema = schema
        self.settings = settings

    def end(self):
        """Extensibility point called after the processing has ended"""

    def deep_transform(self):
        stack = ModelBuilderStack()
        build_stack = [
            TSE(before_transform=True, key=None, data=self.schema.current_model)
        ]
        while build_stack:
            top = build_stack.pop()
            if top.before_transform:
                stack.push(top.key, top.data)
                self.transform_node(stack, top.data)
                build_stack.append(
                    TSE(before_transform=False, key=top.key, data=top.data)
                )
                if isinstance(top.data, dict):
                    # process children in alphabetical order to be deterministic
                    # - need to reverse as pushing into stack
                    for c in reversed(sorted(top.data.items())):
                        build_stack.append(
                            TSE(before_transform=True, key=c[0], data=c[1])
                        )
            else:
                self.after_transform_node(stack, top.data)
                stack.pop()

    def transform_node(self, stack, data):
        """Called on node before children are transformed. To be implemented in inherited class."""

    def after_transform_node(self, stack, data):
        """Called on node after children are transformed. To be implemented in inherited class."""



from .json import JSONOutput


class MappingOutput(JSONOutput):
    TYPE = "mapping"

    def merge_mapping(self, mapping):
        self.stack.merge(mapping)

    def replace_mapping(self, mapping):
        type_ = self.stack.real_top.get("type")
        if type_:
            mapping = {"type": type_, **mapping}
        self.stack.replace(mapping)



from oarepo_model_builder.stack import ModelBuilderStack


def test_stack():
    stack = ModelBuilderStack()

    stack.push("a", {"b": [1, 2]})
    assert stack.path == "/a"
    assert stack.top.key == "a"
    assert stack.top.data == {"b": [1, 2]}
    stack.push("b", [1, 2])
    assert stack.path == "/a/b"
    assert stack.top.key == "b"
    assert stack.top.data == [1, 2]
    stack.pop()
    assert stack.path == "/a"
    assert stack.top.key == "a"
    assert stack.top.data == {"b": [1, 2]}
    stack.pop()



import os

import json5

from oarepo_model_builder.builder import ModelBuilder
from oarepo_model_builder.builders.jsonschema import JSONSchemaBuilder
from oarepo_model_builder.fs import InMemoryFileSystem
from oarepo_model_builder.model_preprocessors.default_values import (
    DefaultValuesModelPreprocessor,
)
from oarepo_model_builder.outputs.jsonschema import JSONSchemaOutput
from oarepo_model_builder.outputs.python import PythonOutput
from oarepo_model_builder.schema import ModelSchema


def test_object():
    data = build_jsonschema(
        {"properties": {"a": {"properties": {"b": {"type": "keyword"}}}}}
    )

    assert data == {
        "type": "object",
        "properties": {
            "a": {"type": "object", "properties": {"b": {"type": "keyword"}}}
        },
    }


def test_array():
    data = build_jsonschema({"properties": {"a": {"items": {"type": "keyword"}}}})

    assert data == {
        "type": "object",
        "properties": {"a": {"type": "array", "items": {"type": "keyword"}}},
    }


def test_object_inside_array():
    data = build_jsonschema(
        {"properties": {"a": {"items": {"properties": {"b": {"type": "keyword"}}}}}}
    )

    assert data == {
        "type": "object",
        "properties": {
            "a": {
                "type": "array",
                "items": {"type": "object", "properties": {"b": {"type": "keyword"}}},
            }
        },
    }


def test_array_brackets():
    data = build_jsonschema(
        {"properties": {"a[]": {"type": "keyword", "^minItems": 5}}}
    )

    assert data == {
        "type": "object",
        "properties": {
            "a": {
                "type": "array",
                "minItems": 5,
                "items": {"type": "keyword"},
            }
        },
    }


def build_jsonschema(model):
    builder = ModelBuilder(
        output_builders=[JSONSchemaBuilder],
        outputs=[JSONSchemaOutput, PythonOutput],
        model_preprocessors=[DefaultValuesModelPreprocessor],
        property_preprocessors=[],
        filesystem=InMemoryFileSystem(),
    )
    builder.build(
        model=ModelSchema(
            "",
            {
                "settings": {
                    "python": {"use-isort": False, "use-black": False},
                },
                "model": {"package": "test", **model},
            },
        ),
        output_dir="",
    )
    return json5.load(
        builder.filesystem.open(
            os.path.join("test", "records", "jsonschemas", "test-1.0.0.json")
        )
    )



from marshmallow import fields

from .model_validation import model_validator
from .utils import ExtendablePartSchema


class RootSchema(ExtendablePartSchema):
    model = fields.Nested(lambda: model_validator.validator_class("model")())
    settings = fields.Nested(lambda: model_validator.validator_class("settings")())
    version = fields.String(required=False, dump_default="1.0.0")
    title = fields.String(required=False)
    output_directory = fields.String(data_key="output-directory", required=False)
    defs = fields.Nested(
        lambda: model_validator.validator_class("properties")(), data_key="$defs"
    )
    runtime_dependencies = fields.Dict(
        fields.String(),
        fields.String(),
        required=False,
        data_key="runtime-dependencies",
    )
    dev_dependencies = fields.Dict(
        fields.String(), fields.String(), required=False, data_key="dev-dependencies"
    )



import os

from oarepo_model_builder.builder import ModelBuilder
from oarepo_model_builder.builders import OutputBuilderComponent
from oarepo_model_builder.builders.jsonschema import JSONSchemaBuilder
from oarepo_model_builder.datatypes import datatypes
from oarepo_model_builder.fs import InMemoryFileSystem
from oarepo_model_builder.model_preprocessors.datatype_default import (
    DatatypeDefaultModelPreprocessor,
)
from oarepo_model_builder.model_preprocessors.default_values import (
    DefaultValuesModelPreprocessor,
)
from oarepo_model_builder.outputs.jsonschema import JSONSchemaOutput
from oarepo_model_builder.outputs.python import PythonOutput
from oarepo_model_builder.schema import ModelSchema
from oarepo_model_builder.validation.model_validation import model_validator
from tests.multilang import MultilangPreprocessor, MultilingualDataType, UIValidator

try:
    import json5
except ImportError:
    import json as json5


def test_simple_jsonschema_builder():
    data = build({"properties": {"a": {"type": "keyword", "ui": {"class": "bolder"}}}})

    assert data == {"type": "object", "properties": {"a": {"type": "keyword"}}}


def test_required():
    data = build(
        {
            "properties": {
                "a": {
                    "type": "keyword",
                    "required": True,
                    "ui": {"class": "bolder"},
                }
            }
        }
    )

    assert data == {
        "type": "object",
        "properties": {"a": {"type": "keyword"}},
    }


def test_required_inside_metadata():
    data = build(
        {
            "properties": {
                "metadata": {
                    "properties": {
                        "a": {
                            "type": "keyword",
                            "required": True,
                            "ui": {"class": "bolder"},
                        }
                    }
                }
            },
            "record-schema-class": "a.BlahSchema",
            "record-ui-schema-class": "a.BlahUISchema",
        }
    )

    assert data == {
        "type": "object",
        "properties": {
            "metadata": {
                "type": "object",
                "properties": {"a": {"type": "keyword"}},
            }
        },
    }


def test_min_length():
    data = build({"properties": {"a": {"type": "keyword", "minLength": 5}}})
    assert data == {
        "type": "object",
        "properties": {"a": {"type": "keyword", "minLength": 5}},
    }


def test_jsonschema_preprocessor():
    data = build(
        {"properties": {"a": {"type": "multilingual", "ui": {"class": "bolder"}}}},
        property_preprocessors=[MultilangPreprocessor],
    )

    assert data == {
        "type": "object",
        "properties": {
            "a": {
                "type": "object",
                "properties": {
                    "lang": {"type": "string"},
                    "value": {"type": "string"},
                },
            }
        },
    }


class TestJSONSchemaOutputComponent(OutputBuilderComponent):
    def model_element_enter(self, builder, data, *, stack):
        if "type" in data:
            data["type"] = "integer"
        return data


def test_components():
    data = build(
        {"properties": {"a": {"type": "keyword", "ui": {"class": "bolder"}}}},
        output_builder_components={
            JSONSchemaOutput.TYPE: [TestJSONSchemaOutputComponent]
        },
    )

    assert data == {"type": "object", "properties": {"a": {"type": "integer"}}}


def build(model, output_builder_components=None, property_preprocessors=None):
    datatypes._prepare_datatypes()
    if UIValidator not in model_validator.validator_map["property-ui"]:
        model_validator.validator_map["property-ui"].append(UIValidator)
    datatypes.datatype_map["multilingual"] = MultilingualDataType
    builder = ModelBuilder(
        output_builders=[JSONSchemaBuilder],
        outputs=[JSONSchemaOutput, PythonOutput],
        model_preprocessors=[
            DefaultValuesModelPreprocessor,
            DatatypeDefaultModelPreprocessor,
        ],
        output_builder_components=output_builder_components,
        filesystem=InMemoryFileSystem(),
        property_preprocessors=property_preprocessors,
    )
    builder.build(
        model=ModelSchema(
            "",
            {
                "settings": {
                    "python": {"use-isort": False, "use-black": False},
                },
                "model": {
                    "package": "test",
                    **model,
                },
            },
        ),
        output_dir="",
    )
    data = json5.load(
        builder.filesystem.open(
            os.path.join("test", "records", "jsonschemas", "test-1.0.0.json")
        )
    )
    return data



import json
import os
import re

from oarepo_model_builder.entrypoints import create_builder_from_entrypoints, load_model
from oarepo_model_builder.fs import InMemoryFileSystem


def test_raw_type():
    schema = load_model(
        "test.yaml",
        "test",
        model_content={
            "model": {"use": "invenio", "properties": {"a": {"type": "flattened"}}},
        },
        isort=False,
        black=False,
    )

    filesystem = InMemoryFileSystem()
    builder = create_builder_from_entrypoints(filesystem=filesystem)

    builder.build(schema, "")

    data = builder.filesystem.open(
        os.path.join("test", "services", "records", "schema.py")
    ).read()
    print(data)
    assert re.sub(r"\s", "", data) == re.sub(
        r"\s",
        "",
        """
from invenio_records_resources.services.records.schema import BaseRecordSchema as InvenioBaseRecordSchema
from marshmallow import ValidationError
from marshmallow import validate as ma_validate
import marshmallow as ma
from marshmallow import fields as ma_fields
from marshmallow_utils import fields as mu_fields
from marshmallow_utils import schemas as mu_schemas
from oarepo_runtime.ui import marshmallow as l10n
from oarepo_runtime.validation import validate_datetime
class TestSchema(InvenioBaseRecordSchema):
    \"""TestSchema schema.\"""
    a = ma_fields.Raw()
    """,
    )

    data = builder.filesystem.read(
        os.path.join("test", "records", "mappings", "os-v2", "test", "test-1.0.0.json")
    )
    data = json.loads(data)

    assert data == {
        "mappings": {
            "properties": {
                "a": {"type": "object", "enabled": False},
                "id": {"type": "keyword"},
                "created": {"type": "date"},
                "updated": {"type": "date"},
                "$schema": {"type": "keyword"},
            },
        }
    }

    data = builder.filesystem.read(
        os.path.join("test", "records", "jsonschemas", "test-1.0.0.json")
    )
    data = json.loads(data)
    print(data)
    assert data == {
        "properties": {
            "a": {"type": "object"},
            "id": {"type": "string"},
            "created": {"type": "string", "format": "date-time"},
            "updated": {"type": "string", "format": "date-time"},
            "$schema": {"type": "string"},
        },
        "type": "object",
    }



import os
import re

from oarepo_model_builder.entrypoints import create_builder_from_entrypoints, load_model
from oarepo_model_builder.fs import InMemoryFileSystem

DUMMY_YAML = "test.yaml"


def test_sort():
    schema = load_model(
        DUMMY_YAML,
        "test",
        model_content={
            "model": {
                "use": "invenio",
                "properties": {
                    "a": {
                        "type": "fulltext+keyword",
                        "sortable": {"key": "a_test"},
                    },
                    "b": {
                        "type": "keyword",
                        "facets": {"field": 'TermsFacet(field="cosi")'},
                        "sortable": {"key": "b_test", "order": "desc"},
                    },
                },
            },
        },
        isort=False,
        black=False,
    )

    filesystem = InMemoryFileSystem()
    builder = create_builder_from_entrypoints(filesystem=filesystem)
    builder.build(schema, "")

    data = builder.filesystem.open(
        os.path.join("test", "services", "records", "search.py")
    ).read()
    assert re.sub(r"\s", "", data) == re.sub(
        r"\s",
        "",
        """
from invenio_records_resources.services import SearchOptions as InvenioSearchOptions
from flask_babelex import lazy_gettext as _
from . import facets

class TestSearchOptions(InvenioSearchOptions):
    \"""TestRecord search options.\"""

    facets = {
    'a_keyword': facets.a_keyword,
    'b': facets.b,
    '_id': facets._id,
    'created': facets.created,
    'updated': facets.updated,
    '_schema': facets._schema,
    }
    sort_options = {
        **InvenioSearchOptions.sort_options,
    'a_test': {'fields': ['a']},
    'b_test': {'fields': ['-b']},
    }
    """,
    )


def test_search_class():
    schema = load_model(
        DUMMY_YAML,
        "test",
        model_content={
            "model": {
                "use": "invenio",
                "properties": {
                    "a": {"type": "fulltext+keyword"},
                    "b": {
                        "type": "keyword",
                        "facets": {"field": 'TermsFacet(field="cosi")'},
                    },
                },
            },
        },
        isort=False,
        black=False,
    )

    filesystem = InMemoryFileSystem()
    builder = create_builder_from_entrypoints(filesystem=filesystem)

    builder.build(schema, "")

    data = builder.filesystem.open(
        os.path.join("test", "services", "records", "search.py")
    ).read()
    print(data)
    assert re.sub(r"\s", "", data) == re.sub(
        r"\s",
        "",
        """
from invenio_records_resources.services import SearchOptions as InvenioSearchOptions
from flask_babelex import lazy_gettext as _
from . import facets


class TestSearchOptions(InvenioSearchOptions):
    \"""TestRecord search options.\"""

    facets = {
      'a_keyword': facets.a_keyword,
      'b': facets.b,
      '_id': facets._id,
      'created': facets.created,
      'updated': facets.updated,
      '_schema': facets._schema,
    }
    sort_options = {
        **InvenioSearchOptions.sort_options,
    }
    """,
    )



from pathlib import Path
from typing import Union

from oarepo_model_builder.builder import ModelBuilder
from oarepo_model_builder.schema import ModelSchema


class Profile:
    def build(
        self,
        model: ModelSchema,
        output_directory: Union[str, Path],
        builder: ModelBuilder,
        **kwargs,
    ):
        builder.build(model, output_directory, **kwargs)



import os
import re

from oarepo_model_builder.entrypoints import create_builder_from_entrypoints, load_model
from oarepo_model_builder.fs import InMemoryFileSystem

DUMMY_YAML = "test.yaml"


def test_include_invenio():
    schema = load_model(
        DUMMY_YAML,
        "test",
        model_content={
            "model": {
                "use": "invenio",
                "properties": {
                    "jej": {
                        "type": "nested",
                        "properties": {
                            "c": {
                                "type": "keyword",
                                "facets": {"field": 'TermsFacet(field="cosi")'},
                            }
                        },
                    },
                    "a": "fulltext+keyword",
                    "b": {
                        "type": "keyword",
                        "facets": {"field": 'TermsFacet(field="cosi")'},
                    },
                    "c": "fulltext",
                },
            },
        },
        isort=False,
        black=False,
    )

    filesystem = InMemoryFileSystem()
    builder = create_builder_from_entrypoints(filesystem=filesystem)
    builder.build(schema, "")

    data = builder.filesystem.open(
        os.path.join("test", "services", "records", "facets.py")
    ).read()
    assert re.sub(r"\s", "", data) == re.sub(
        r"\s",
        "",
        """
\"""Facet definitions.\"""

from invenio_search.engine import dsl
from flask_babelex import lazy_gettext as _

from invenio_records_resources.services.records.facets import TermsFacet
from oarepo_runtime.facets.date import DateTimeFacet
from oarepo_runtime.facets.nested_facet import NestedLabeledFacet



jej_c = NestedLabeledFacet(path ="jej", nested_facet = TermsFacet(field="cosi"))



a_keyword = TermsFacet(field="a.keyword", label=_("a/keyword.label") )



b = TermsFacet(field="cosi")



_id = TermsFacet(field="id", label=_("id.label") )



created = DateTimeFacet(field="created", label=_("created.label") )



updated = DateTimeFacet(field="updated", label=_("updated.label") )



_schema = TermsFacet(field="$schema", label=_("$schema.label") )


    """,
    )


def test_nested():
    schema = load_model(
        DUMMY_YAML,
        "test",
        model_content={
            "model": {
                "use": "invenio",
                "properties": {
                    "b": {
                        "type": "nested",
                        "properties": {
                            "c": {
                                "type": "keyword",
                            },
                            "d": {"type": "fulltext+keyword"},
                            "f": {
                                "type": "nested",
                                "properties": {"g": {"type": "keyword"}},
                            },
                        },
                    }
                },
            },
        },
        isort=False,
        black=False,
    )

    filesystem = InMemoryFileSystem()
    builder = create_builder_from_entrypoints(filesystem=filesystem)

    builder.build(schema, "")

    data = builder.filesystem.open(
        os.path.join("test", "services", "records", "facets.py")
    ).read()
    print(data)
    assert re.sub(r"\s", "", data) == re.sub(
        r"\s",
        "",
        """
\"""Facet definitions.\"""

from invenio_search.engine import dsl
from flask_babelex import lazy_gettext as _

from invenio_records_resources.services.records.facets import TermsFacet
from oarepo_runtime.facets.date import DateTimeFacet
from oarepo_runtime.facets.nested_facet import NestedLabeledFacet



b_c = NestedLabeledFacet(path ="b", nested_facet = TermsFacet(field="b.c", label=_("b/c.label") ))



b_d_keyword = NestedLabeledFacet(path ="b", nested_facet = TermsFacet(field="b.d.keyword", label=_("b/d/keyword.label") ))



b_f_g = NestedLabeledFacet(path ="b", nested_facet = NestedLabeledFacet(path ="b.f", nested_facet = TermsFacet(field="b.f.g", label=_("b/f/g.label") )))



_id = TermsFacet(field="id", label=_("id.label") )



created = DateTimeFacet(field="created", label=_("created.label") )



updated = DateTimeFacet(field="updated", label=_("updated.label") )



_schema = TermsFacet(field="$schema", label=_("$schema.label") )

""",
    )


def test_object():
    schema = load_model(
        DUMMY_YAML,
        "test",
        model_content={
            "model": {
                "use": "invenio",
                "properties": {
                    "b": {
                        "type": "object",
                        "properties": {
                            "c": {
                                "type": "keyword",
                            },
                            "d": {"type": "fulltext+keyword"},
                            "f": {
                                "type": "object",
                                "properties": {"g": {"type": "keyword"}},
                            },
                            "e": "fulltext",
                        },
                    }
                },
            },
        },
        isort=False,
        black=False,
    )

    filesystem = InMemoryFileSystem()
    builder = create_builder_from_entrypoints(filesystem=filesystem)

    builder.build(schema, "")

    data = builder.filesystem.open(
        os.path.join("test", "services", "records", "facets.py")
    ).read()
    assert re.sub(r"\s", "", data) == re.sub(
        r"\s",
        "",
        """
        \"""Facet definitions.\"""

from invenio_search.engine import dsl
from flask_babelex import lazy_gettext as _

from invenio_records_resources.services.records.facets import TermsFacet
from oarepo_runtime.facets.date import DateTimeFacet



b_c = TermsFacet(field="b.c", label=_("b/c.label") )



b_d_keyword = TermsFacet(field="b.d.keyword", label=_("b/d/keyword.label") )



b_f_g = TermsFacet(field="b.f.g", label=_("b/f/g.label") )



_id = TermsFacet(field="id", label=_("id.label") )



created = DateTimeFacet(field="created", label=_("created.label") )



updated = DateTimeFacet(field="updated", label=_("updated.label") )



_schema = TermsFacet(field="$schema", label=_("$schema.label") )
""",
    )


def test_nest_obj():
    schema = load_model(
        DUMMY_YAML,
        "test",
        model_content={
            "model": {
                "use": "invenio",
                "properties": {
                    "b_nes": {
                        "type": "nested",
                        "properties": {
                            "c": {
                                "type": "keyword",
                            },
                            "d": {"type": "fulltext+keyword"},
                            "f": {
                                "type": "object",
                                "properties": {"g": {"type": "keyword"}},
                            },
                        },
                    },
                    "b_obj": {
                        "type": "object",
                        "properties": {
                            "c": {
                                "type": "keyword",
                            },
                            "d": {"type": "fulltext+keyword"},
                            "f": {
                                "type": "nested",
                                "properties": {"g": {"type": "keyword"}},
                            },
                        },
                    },
                },
            },
        },
        isort=False,
        black=False,
    )

    filesystem = InMemoryFileSystem()
    builder = create_builder_from_entrypoints(filesystem=filesystem)

    builder.build(schema, "")

    data = builder.filesystem.open(
        os.path.join("test", "services", "records", "facets.py")
    ).read()
    assert re.sub(r"\s", "", data) == re.sub(
        r"\s",
        "",
        """
        \"""Facet definitions.\"""

from invenio_search.engine import dsl
from flask_babelex import lazy_gettext as _

from invenio_records_resources.services.records.facets import TermsFacet
from oarepo_runtime.facets.date import DateTimeFacet
from oarepo_runtime.facets.nested_facet import NestedLabeledFacet



b_nes_c = NestedLabeledFacet(path ="b_nes", nested_facet = TermsFacet(field="b_nes.c", label=_("b_nes/c.label") ))



b_nes_d_keyword = NestedLabeledFacet(path ="b_nes", nested_facet = TermsFacet(field="b_nes.d.keyword", label=_("b_nes/d/keyword.label") ))



b_nes_f_g = NestedLabeledFacet(path ="b_nes", nested_facet = TermsFacet(field="b_nes.f.g", label=_("b_nes/f/g.label") ))



b_obj_c = TermsFacet(field="b_obj.c", label=_("b_obj/c.label") )



b_obj_d_keyword = TermsFacet(field="b_obj.d.keyword", label=_("b_obj/d/keyword.label") )



b_obj_f_g = NestedLabeledFacet(path ="b_obj.f", nested_facet = TermsFacet(field="b_obj.f.g", label=_("b_obj/f/g.label") ))



_id = TermsFacet(field="id", label=_("id.label") )



created = DateTimeFacet(field="created", label=_("created.label") )



updated = DateTimeFacet(field="updated", label=_("updated.label") )



_schema = TermsFacet(field="$schema", label=_("$schema.label") )
""",
    )


def test_array():
    schema = load_model(
        DUMMY_YAML,
        "test",
        model_content={
            "model": {
                "use": "invenio",
                "properties": {
                    "a[]": "keyword",
                    "b[]": "fulltext",
                    "c[]": "fulltext+keyword",
                },
            },
        },
        isort=False,
        black=False,
    )

    filesystem = InMemoryFileSystem()
    builder = create_builder_from_entrypoints(filesystem=filesystem)

    builder.build(schema, "")
    data = builder.filesystem.open(
        os.path.join("test", "services", "records", "facets.py")
    ).read()
    print(data)

    assert re.sub(r"\s", "", data) == re.sub(
        r"\s",
        "",
        """
        \"""Facet definitions.\"""

from invenio_search.engine import dsl
from flask_babelex import lazy_gettext as _

from invenio_records_resources.services.records.facets import TermsFacet
from oarepo_runtime.facets.date import DateTimeFacet



a = TermsFacet(field="a", label=_("a.label") )



c_keyword = TermsFacet(field="c.keyword", label=_("c/keyword.label") )



_id = TermsFacet(field="id", label=_("id.label") )



created = DateTimeFacet(field="created", label=_("created.label") )



updated = DateTimeFacet(field="updated", label=_("updated.label") )



_schema = TermsFacet(field="$schema", label=_("$schema.label") )
""",
    )


def test_array_object():
    schema = load_model(
        DUMMY_YAML,
        "test",
        model_content={
            "model": {
                "use": "invenio",
                "properties": {
                    "arr": {
                        "type": "array",
                        "items": {
                            "type": "object",
                            "properties": {
                                "a": {
                                    "type": "array",
                                    "items": {
                                        "type": "object",
                                        "properties": {"c": "keyword"},
                                    },
                                }
                            },
                        },
                    },
                    "test": {
                        "type": "array",
                        "items": {"type": "array", "items": {"type": "keyword"}},
                    },
                    "test2": {
                        "type": "array",
                        "items": {
                            "type": "nested",
                            "properties": {"g": {"type": "keyword"}},
                        },
                    },
                },
            },
        },
        isort=False,
        black=False,
    )

    filesystem = InMemoryFileSystem()
    builder = create_builder_from_entrypoints(filesystem=filesystem)

    builder.build(schema, "")

    data = builder.filesystem.open(
        os.path.join("test", "services", "records", "facets.py")
    ).read()
    print(data)
    assert re.sub(r"\s", "", data) == re.sub(
        r"\s",
        "",
        """
 \"""Facet definitions.\"""

from invenio_search.engine import dsl
from flask_babelex import lazy_gettext as _

from invenio_records_resources.services.records.facets import TermsFacet
from oarepo_runtime.facets.date import DateTimeFacet
from oarepo_runtime.facets.nested_facet import NestedLabeledFacet



arr_a_c = TermsFacet(field="arr.a.c", label=_("arr/a/c.label") )



test = TermsFacet(field="test", label=_("test.label") )



test2_g = NestedLabeledFacet(path ="test2", nested_facet = TermsFacet(field="test2.g", label=_("test2/g.label") ))



_id = TermsFacet(field="id", label=_("id.label") )



created = DateTimeFacet(field="created", label=_("created.label") )



updated = DateTimeFacet(field="updated", label=_("updated.label") )



_schema = TermsFacet(field="$schema", label=_("$schema.label") )
""",
    )


def test_array_nested():
    schema = load_model(
        DUMMY_YAML,
        "test",
        model_content={
            "model": {
                "use": "invenio",
                "properties": {
                    "obj": {
                        "type": "object",
                        "properties": {
                            "arr": {
                                "type": "array",
                                "items": {
                                    "type": "nested",
                                    "properties": {
                                        "d": {
                                            "type": "fulltext+keyword",
                                            # "facets": {"key": "test"},
                                        },
                                        "e": {
                                            "type": "object",
                                            "properties": {"f": "keyword"},
                                        },
                                    },
                                },
                            }
                        },
                    },
                },
            },
        },
        isort=False,
        black=False,
    )

    filesystem = InMemoryFileSystem()
    builder = create_builder_from_entrypoints(filesystem=filesystem)

    builder.build(schema, "")

    data = builder.filesystem.open(
        os.path.join("test", "services", "records", "facets.py")
    ).read()
    assert re.sub(r"\s", "", data) == re.sub(
        r"\s",
        "",
        """
        \"""Facet definitions.\"""

from invenio_search.engine import dsl
from flask_babelex import lazy_gettext as _

from invenio_records_resources.services.records.facets import TermsFacet
from oarepo_runtime.facets.date import DateTimeFacet
from oarepo_runtime.facets.nested_facet import NestedLabeledFacet



obj_arr_d_keyword = NestedLabeledFacet(path ="obj.arr", nested_facet = TermsFacet(field="obj.arr.d.keyword", label=_("obj/arr/d/keyword.label") ))



obj_arr_e_f = NestedLabeledFacet(path ="obj.arr", nested_facet = TermsFacet(field="obj.arr.e.f", label=_("obj/arr/e/f.label") ))



_id = TermsFacet(field="id", label=_("id.label") )



created = DateTimeFacet(field="created", label=_("created.label") )



updated = DateTimeFacet(field="updated", label=_("updated.label") )



_schema = TermsFacet(field="$schema", label=_("$schema.label") )
""",
    )


def test_not_searchable():
    schema = load_model(
        DUMMY_YAML,
        "test",
        model_content={
            "model": {
                "use": "invenio",
                "properties": {
                    "a": {"type": "fulltext+keyword", "facets": {"searchable": False}},
                    "b": {
                        "type": "keyword",
                        "facets": {"field": 'TermsFacet(field="cosi")'},
                    },
                    "arr": {
                        "type": "array",
                        "items": {
                            "type": "nested",
                            "properties": {
                                "d": {
                                    "type": "keyword",
                                    "facets": {"searchable": False},
                                },
                                "e": {
                                    "type": "object",
                                    "properties": {
                                        "f": "keyword",
                                    },
                                },
                            },
                        },
                    },
                },
            },
        },
        isort=False,
        black=False,
    )

    filesystem = InMemoryFileSystem()
    builder = create_builder_from_entrypoints(filesystem=filesystem)
    builder.build(schema, "")

    data = builder.filesystem.open(
        os.path.join("test", "services", "records", "facets.py")
    ).read()

    assert re.sub(r"\s", "", data) == re.sub(
        r"\s",
        "",
        """
\"""Facet definitions.\"""

from invenio_search.engine import dsl
from flask_babelex import lazy_gettext as _

from invenio_records_resources.services.records.facets import TermsFacet
from oarepo_runtime.facets.date import DateTimeFacet
from oarepo_runtime.facets.nested_facet import NestedLabeledFacet



b = TermsFacet(field="cosi")



arr_e_f = NestedLabeledFacet(path ="arr", nested_facet = TermsFacet(field="arr.e.f", label=_("arr/e/f.label") ))



_id = TermsFacet(field="id", label=_("id.label") )



created = DateTimeFacet(field="created", label=_("created.label") )



updated = DateTimeFacet(field="updated", label=_("updated.label") )



_schema = TermsFacet(field="$schema", label=_("$schema.label") )

    """,
    )


def test_top_facets():
    schema = load_model(
        DUMMY_YAML,
        "test",
        model_content={
            "model": {
                "use": "invenio",
                "searchable": False,
                "properties": {
                    "a": {"type": "fulltext+keyword", "facets": {"searchable": True}},
                    "b": {
                        "type": "keyword",
                        "facets": {"field": 'TermsFacet(field="cosi")'},
                    },
                    "c": "keyword",
                    "arr": {
                        "type": "array",
                        "facets": {"searchable": True},
                        "items": {
                            "type": "nested",
                            "properties": {
                                "d": {"type": "keyword"},
                                "e": {
                                    "type": "object",
                                    "properties": {
                                        "f": "keyword",
                                    },
                                },
                            },
                        },
                    },
                    "lst[]": "keyword",
                },
            },
        },
        isort=False,
        black=False,
    )

    filesystem = InMemoryFileSystem()
    builder = create_builder_from_entrypoints(filesystem=filesystem)
    builder.build(schema, "")

    data = builder.filesystem.open(
        os.path.join("test", "services", "records", "facets.py")
    ).read()

    assert re.sub(r"\s", "", data) == re.sub(
        r"\s",
        "",
        """
\"""Facet definitions.\"""

from invenio_search.engine import dsl
from flask_babelex import lazy_gettext as _

from invenio_records_resources.services.records.facets import TermsFacet
from oarepo_runtime.facets.date import DateTimeFacet
from oarepo_runtime.facets.nested_facet import NestedLabeledFacet



a_keyword = TermsFacet(field="a.keyword", label=_("a/keyword.label") )



b = TermsFacet(field="cosi")



arr_d = NestedLabeledFacet(path ="arr", nested_facet = TermsFacet(field="arr.d", label=_("arr/d.label") ))



arr_e_f = NestedLabeledFacet(path ="arr", nested_facet = TermsFacet(field="arr.e.f", label=_("arr/e/f.label") ))



_id = TermsFacet(field="id", label=_("id.label") )



created = DateTimeFacet(field="created", label=_("created.label") )



updated = DateTimeFacet(field="updated", label=_("updated.label") )



_schema = TermsFacet(field="$schema", label=_("$schema.label") )

    """,
    )


def test_searchable_true():
    schema = load_model(
        DUMMY_YAML,
        "test",
        model_content={
            "model": {
                "use": "invenio",
                "properties": {
                    "a": {"type": "fulltext+keyword", "facets": {"searchable": False}},
                    "b": {
                        "type": "keyword",
                        "facets": {"field": 'TermsFacet(field="cosi")'},
                    },
                    "c": "fulltext",
                    "f": {
                        "type": "object",
                        "facets": {"searchable": False},
                        "properties": {"g": {"type": "keyword"}},
                    },
                },
            },
        },
        isort=False,
        black=False,
    )

    filesystem = InMemoryFileSystem()
    builder = create_builder_from_entrypoints(filesystem=filesystem)
    builder.build(schema, "")
    data = builder.filesystem.open(
        os.path.join("test", "services", "records", "facets.py")
    ).read()

    assert re.sub(r"\s", "", data) == re.sub(
        r"\s",
        "",
        """
\"""Facet definitions.\"""

from invenio_search.engine import dsl
from flask_babelex import lazy_gettext as _

from invenio_records_resources.services.records.facets import TermsFacet
from oarepo_runtime.facets.date import DateTimeFacet



b = TermsFacet(field="cosi")



_id = TermsFacet(field="id", label=_("id.label") )



created = DateTimeFacet(field="created", label=_("created.label") )



updated = DateTimeFacet(field="updated", label=_("updated.label") )



_schema = TermsFacet(field="$schema", label=_("$schema.label") )

    """,
    )


def test_enum():
    schema = load_model(
        DUMMY_YAML,
        "test",
        model_content={
            "model": {
                "use": "invenio",
                "properties": {
                    "a": {"type": "keyword", "enum": ["a", "b"]},
                    "b": {
                        "type": "keyword",
                    },
                },
            },
        },
        isort=False,
        black=False,
    )

    filesystem = InMemoryFileSystem()
    builder = create_builder_from_entrypoints(filesystem=filesystem)
    builder.build(schema, "")

    data = builder.filesystem.open(
        os.path.join("test", "services", "records", "facets.py")
    ).read()

    assert re.sub(r"\s", "", data) == re.sub(
        r"\s",
        "",
        """
\"""Facet definitions.\"""

from invenio_search.engine import dsl
from flask_babelex import lazy_gettext as _

from invenio_records_resources.services.records.facets import TermsFacet
from oarepo_runtime.facets.date import DateTimeFacet
from oarepo_runtime.facets.enum import EnumTermsFacet


a = EnumTermsFacet(field="a", label=_("a.label") )



b = TermsFacet(field="b", label=_("b.label") )


_id = TermsFacet(field="id", label=_("id.label") )



created = DateTimeFacet(field="created", label=_("created.label") )



updated = DateTimeFacet(field="updated", label=_("updated.label") )



_schema = TermsFacet(field="$schema", label=_("$schema.label") )

    """,
    )


def test_customizations_args_class():
    schema = load_model(
        DUMMY_YAML,
        "test",
        model_content={
            "model": {
                "use": "invenio",
                "properties": {
                    "a": {
                        "type": "keyword",
                        "facets": {
                            "facet-class": "MyFacetClass",
                            "args": ["blah=123"],
                            "imports": [{"import": "blah.MyFacetClass"}],
                        },
                    },
                },
            },
        },
        isort=False,
        black=False,
    )

    filesystem = InMemoryFileSystem()
    builder = create_builder_from_entrypoints(filesystem=filesystem)
    builder.build(schema, "")

    data = builder.filesystem.open(
        os.path.join("test", "services", "records", "facets.py")
    ).read()
    print(data)

    assert re.sub(r"\s", "", data) == re.sub(
        r"\s",
        "",
        """
\"""Facet definitions.\"""

from invenio_search.engine import dsl
from flask_babelex import lazy_gettext as _

from blah import MyFacetClass
from invenio_records_resources.services.records.facets import TermsFacet
from oarepo_runtime.facets.date import DateTimeFacet


a = MyFacetClass(field="a", label=_("a.label"), blah=123 )


_id = TermsFacet(field="id", label=_("id.label") )



created = DateTimeFacet(field="created", label=_("created.label") )



updated = DateTimeFacet(field="updated", label=_("updated.label") )



_schema = TermsFacet(field="$schema", label=_("$schema.label") )

    """,
    )


def test_customizations_field():
    schema = load_model(
        DUMMY_YAML,
        "test",
        model_content={
            "model": {
                "use": "invenio",
                "properties": {
                    "a": {
                        "type": "keyword",
                        "facets": {
                            "facet-class": "MyFacetClass",
                            "args": ["blah=123"],
                            "path": "aaa",
                            "imports": [{"import": "blah.MyFacetClass"}],
                        },
                    },
                },
            },
        },
        isort=False,
        black=False,
    )

    filesystem = InMemoryFileSystem()
    builder = create_builder_from_entrypoints(filesystem=filesystem)
    builder.build(schema, "")

    data = builder.filesystem.open(
        os.path.join("test", "services", "records", "facets.py")
    ).read()
    print(data)
    assert re.sub(r"\s", "", data) == re.sub(
        r"\s",
        "",
        """
\"""Facet definitions.\"""

from invenio_search.engine import dsl
from flask_babelex import lazy_gettext as _

from blah import MyFacetClass

from invenio_records_resources.services.records.facets import TermsFacet

from oarepo_runtime.facets.date import DateTimeFacet


a = MyFacetClass(field="a.aaa", label=_("a.label"), blah=123 )


_id = TermsFacet(field="id", label=_("id.label") )



created = DateTimeFacet(field="created", label=_("created.label") )



updated = DateTimeFacet(field="updated", label=_("updated.label") )



_schema = TermsFacet(field="$schema", label=_("$schema.label") )

    """,
    )



from oarepo_model_builder.builders.extend import ExtendBuilder
from oarepo_model_builder.property_preprocessors import PropertyPreprocessor, process
from oarepo_model_builder.stack import ModelBuilderStack


class DisableMarshmallowPreprocessor(PropertyPreprocessor):
    TYPE = "disable-marshmallow"

    @process(
        model_builder=ExtendBuilder,
        path="/properties/**",
        condition=lambda current, stack: stack.top.schema_valid,
    )
    def modify_marshmallow(self, data, stack: ModelBuilderStack, **kwargs):
        if stack.top.schema_element_type in ("property", "items"):
            self._process_marshmallow_def(data.setdefault("marshmallow", {}))
            self._process_marshmallow_def(
                data.setdefault("ui", {}).setdefault("marshmallow", {})
            )

    def _process_marshmallow_def(self, marshmallow):
        if not marshmallow:
            marshmallow.update({"read": False, "write": False})
        elif "schema-class" in marshmallow:
            # already generated - if user wants to override this, he has to set schema-class and generate
            marshmallow["generate"] = False



from typing import Mapping

import munch


class HyphenMunch(munch.AutoMunch):
    def __setitem__(self, key, value):
        if isinstance(value, Mapping) and not isinstance(
            value, (munch.AutoMunch, munch.Munch)
        ):
            value = munch.munchify(value, HyphenMunch)

        return super().__setitem__(key, value)

    def __getitem__(self, key):
        try:
            return super().__getitem__(key)
        except:
            key = key.replace("_", "-")
            return super().__getitem__(key)

    def get(self, key, default=None):
        try:
            return self.__getitem__(key)
        except KeyError:
            return default

    def __setattr__(self, k, v):
        k = k.replace("_", "-")
        self[k] = v



from marshmallow import fields

from oarepo_model_builder.validation.property_marshmallow import ImportSchema

from .utils import ExtendablePartSchema


class PropertyFacets(ExtendablePartSchema):
    key = fields.String(required=False)
    field = fields.String(required=False)
    searchable = fields.Boolean(required=False)
    facet_class = fields.String(required=False, data_key="facet-class")
    args = fields.List(fields.String(), required=False)
    imports = fields.List(fields.Nested(ImportSchema), required=False)
    path = fields.String(required=False)



import os
import re
import sys

import pytest

from oarepo_model_builder.builder import ModelBuilder
from oarepo_model_builder.fs import InMemoryFileSystem
from oarepo_model_builder.invenio.invenio_record_resource_config import (
    InvenioRecordResourceConfigBuilder,
)
from oarepo_model_builder.invenio.invenio_record_ui_schema import (
    InvenioRecordUISchemaBuilder,
)
from oarepo_model_builder.invenio.invenio_record_ui_serializer import (
    InvenioRecordUISerializerBuilder,
)
from oarepo_model_builder.model_preprocessors.datatype_default import (
    DatatypeDefaultModelPreprocessor,
)
from oarepo_model_builder.model_preprocessors.default_values import (
    DefaultValuesModelPreprocessor,
)
from oarepo_model_builder.model_preprocessors.invenio import InvenioModelPreprocessor
from oarepo_model_builder.model_preprocessors.opensearch import (
    OpensearchModelPreprocessor,
)
from oarepo_model_builder.outputs.python import PythonOutput
from oarepo_model_builder.property_preprocessors.datatype_preprocessor import (
    DataTypePreprocessor,
)
from oarepo_model_builder.schema import ModelSchema

OAREPO_MARSHMALLOW = "marshmallow"
B_SCHEMA = 'classB(ma.Schema):"""Bschema."""b=ma_fields.String()'
BUI_SCHEMA = 'classBUISchema(ma.Schema):"""BUISchemaschema."""b=ma_fields.String()'


def get_test_schema(**props):
    return ModelSchema(
        "",
        {
            "settings": {
                "python": {"use-isort": False, "use-black": False},
            },
            "model": {"package": "test", "properties": props},
        },
    )


@pytest.fixture
def fulltext_builder():
    return ModelBuilder(
        output_builders=[
            InvenioRecordUISchemaBuilder,
            InvenioRecordUISerializerBuilder,
            InvenioRecordResourceConfigBuilder,
        ],
        outputs=[PythonOutput],
        model_preprocessors=[
            DefaultValuesModelPreprocessor,
            OpensearchModelPreprocessor,
            InvenioModelPreprocessor,
            DatatypeDefaultModelPreprocessor,
        ],
        property_preprocessors=[DataTypePreprocessor],
    )


def _test(fulltext_builder, string_type):
    schema = get_test_schema(a={"type": string_type})
    fulltext_builder.filesystem = InMemoryFileSystem()
    fulltext_builder.build(schema, output_dir="")

    with fulltext_builder.filesystem.open(
        os.path.join("test", "services", "records", "ui_schema.py")  # NOSONAR
    ) as f:
        data = f.read()
    assert "a = ma_fields.String()" in data


def test_fulltext(fulltext_builder):
    _test(fulltext_builder, "fulltext")


def test_keyword(fulltext_builder):
    _test(fulltext_builder, "keyword")


def test_fulltext_keyword(fulltext_builder):
    _test(fulltext_builder, "fulltext+keyword")


def test_simple_array(fulltext_builder):
    schema = get_test_schema(a={"type": "array", "items": {"type": "keyword"}})
    fulltext_builder.filesystem = InMemoryFileSystem()
    fulltext_builder.build(schema, output_dir="")

    with fulltext_builder.filesystem.open(
        os.path.join("test", "services", "records", "ui_schema.py")
    ) as f:
        data = f.read()
    assert "a = ma_fields.List(ma_fields.String())" in data


def test_array_of_objects(fulltext_builder):
    print("starting test ...", file=sys.stderr)
    schema = get_test_schema(
        a={
            "type": "array",
            "items": {"type": "object", "properties": {"b": {"type": "integer"}}},
        }
    )
    fulltext_builder.filesystem = InMemoryFileSystem()
    fulltext_builder.build(schema, output_dir="")

    with fulltext_builder.filesystem.open(
        os.path.join("test", "services", "records", "ui_schema.py")
    ) as f:
        data = f.read()
    assert (
        'class AItemUISchema(ma.Schema):\n    """AItemUISchema schema."""\n    b = ma_fields.Integer()'
        in data
    )
    assert "a = ma_fields.List(ma_fields.Nested(lambda: AItemUISchema()))" in data


def test_generate_nested_schema_same_file(fulltext_builder):
    schema = get_test_schema(
        a={
            "type": "object",
            OAREPO_MARSHMALLOW: {"schema-class": "B", "generate": True},
            "ui": {OAREPO_MARSHMALLOW: {"schema-class": "BUISchema", "generate": True}},
            "properties": {
                "b": {
                    "type": "keyword",
                }
            },
        }
    )
    fulltext_builder.filesystem = InMemoryFileSystem()
    fulltext_builder.build(schema, output_dir="")

    with fulltext_builder.filesystem.open(
        os.path.join("test", "services", "records", "ui_schema.py")
    ) as f:
        data = f.read()
    assert (
        re.sub(
            r"\s",
            "",
            """class BUISchema(ma.Schema):
        \"""BUISchema schema.\"""
        
        b = ma_fields.String()""",
        )
        in re.sub(r"\s", "", data)
    )
    assert (
        re.sub(
            r"\s",
            "",
            """class TestUISchema(ma.Schema):
        \"""TestUISchema schema.\"""
        
        a = ma_fields.Nested(lambda: BUISchema())""",
        )
        in re.sub(r"\s", "", data)
    )


def test_generate_nested_schema_different_file(fulltext_builder):
    schema = get_test_schema(
        a={
            "type": "object",
            OAREPO_MARSHMALLOW: {
                "schema-class": "test.services.schema2.B",
                "generate": True,
            },
            "ui": {
                OAREPO_MARSHMALLOW: {
                    "schema-class": "test.services.schema2.BUISchema",
                    "generate": True,
                }
            },
            "properties": {
                "b": {
                    "type": "keyword",
                }
            },
        }
    )
    fulltext_builder.filesystem = InMemoryFileSystem()
    fulltext_builder.build(schema, output_dir="")

    with fulltext_builder.filesystem.open(
        os.path.join("test", "services", "records", "ui_schema.py")
    ) as f:
        data = f.read()

    assert (
        re.sub(
            r"\s",
            "",
            """class TestUISchema(ma.Schema):
        \"""TestUISchema schema.\"""
    
        a = ma_fields.Nested(lambda: BUISchema())""",
        )
        in re.sub(r"\s", "", data)
    )


def test_use_nested_schema_same_file(fulltext_builder):
    schema = get_test_schema(
        a={
            "type": "object",
            OAREPO_MARSHMALLOW: {"schema-class": "B", "generate": False},
            "ui": {
                OAREPO_MARSHMALLOW: {"schema-class": "BUISchema", "generate": False}
            },
            "properties": {
                "b": {
                    "type": "keyword",
                }
            },
        }
    )
    fulltext_builder.filesystem = InMemoryFileSystem()
    fulltext_builder.build(schema, output_dir="")

    with fulltext_builder.filesystem.open(
        os.path.join("test", "services", "records", "ui_schema.py")
    ) as f:
        data = f.read()
        print(data)
    assert (
        re.sub(
            r"\s",
            "",
            """class TestUISchema(ma.Schema):
        \"""TestUISchema schema.\"""
        
        a = ma_fields.Nested(lambda: BUISchema())""",
        )
        in re.sub(r"\s", "", data)
    )
    assert "classBUISchema(ma.Schema)" not in re.sub(r"\s", "", data)


def test_use_nested_schema_different_file(fulltext_builder):
    schema = get_test_schema(
        a={
            "type": "object",
            OAREPO_MARSHMALLOW: {"schema-class": "c.B", "generate": False},
            "ui": {
                OAREPO_MARSHMALLOW: {"schema-class": "c.BUISchema", "generate": False}
            },
            "properties": {
                "b": {
                    "type": "keyword",
                }
            },
        }
    )
    fulltext_builder.filesystem = InMemoryFileSystem()
    fulltext_builder.build(schema, output_dir="")

    with fulltext_builder.filesystem.open(
        os.path.join("test", "services", "records", "ui_schema.py")
    ) as f:
        data = f.read()
    assert re.sub(r"\s", "", "from c import BUISchema") in re.sub(r"\s", "", data)
    assert (
        'classTestUISchema(ma.Schema):"""TestUISchemaschema."""a=ma_fields.Nested(lambda:BUISchema())'
        in re.sub(r"\s", "", data)
    )


def test_generate_nested_schema_array(fulltext_builder):
    schema = get_test_schema(
        a={
            "type": "array",
            "items": {
                "type": "object",
                OAREPO_MARSHMALLOW: {"schema-class": "B", "generate": True},
                "ui": {
                    OAREPO_MARSHMALLOW: {"schema-class": "BUISchema", "generate": True}
                },
                "properties": {
                    "b": {
                        "type": "keyword",
                    }
                },
            },
        }
    )
    fulltext_builder.filesystem = InMemoryFileSystem()
    fulltext_builder.build(schema, output_dir="")

    with fulltext_builder.filesystem.open(
        os.path.join("test", "services", "records", "ui_schema.py")
    ) as f:
        data = f.read()
    assert BUI_SCHEMA in re.sub(r"\s", "", data)
    assert (
        'classTestUISchema(ma.Schema):"""TestUISchemaschema."""a=ma_fields.List(ma_fields.Nested(lambda:BUISchema()))'
        in re.sub(r"\s", "", data)
    )


def test_extend_existing(fulltext_builder):
    schema = get_test_schema(a={"type": "keyword"}, b={"type": "keyword"})
    fulltext_builder.filesystem = InMemoryFileSystem()

    with fulltext_builder.filesystem.open(
        os.path.join("test", "services", "records", "ui_schema.py"), "w"
    ) as f:
        f.write(
            '''
from invenio_records_resources.services.records.schema import BaseRecordSchema as InvenioBaseRecordSchema
import marshmallow as ma
import marshmallow.fields as ma_fields
import marshmallow.validate as ma_valid
class TestUISchema(ma.Schema):
    """TestUISchema schema."""
    a = ma_fields.String()'''
        )

    fulltext_builder.build(schema, output_dir="")
    data = fulltext_builder.filesystem.read(
        os.path.join("test", "services", "records", "ui_schema.py")
    )
    assert "b = ma_fields.String()" in data


def test_generate_nested_schema_relative_same_package(fulltext_builder):
    schema = get_test_schema(
        a={
            "type": "object",
            OAREPO_MARSHMALLOW: {
                "schema-class": ".schema2.B",
                "generate": True,
            },
            "ui": {
                OAREPO_MARSHMALLOW: {
                    "schema-class": ".ui_schema2.BUISchema",
                    "generate": True,
                }
            },
            "properties": {
                "b": {
                    "type": "keyword",
                }
            },
        }
    )
    fulltext_builder.filesystem = InMemoryFileSystem()
    fulltext_builder.build(schema, output_dir="")

    with fulltext_builder.filesystem.open(
        os.path.join("test", "services", "records", "ui_schema.py")
    ) as f:
        data = f.read()

    assert (
        re.sub(
            r"\s",
            "",
            """class TestUISchema(ma.Schema):
        \"""TestUISchema schema.\"""
    
        a = ma_fields.Nested(lambda:BUISchema())""",
        )
        in re.sub(r"\s", "", data)
    )

    assert "from test.services.records.ui_schema2 import BUISchema" in data
    with fulltext_builder.filesystem.open(
        os.path.join("test", "services", "records", "ui_schema2.py")
    ) as f:
        data = f.read()
    assert BUI_SCHEMA in re.sub(r"\s", "", data)


def test_generate_nested_schema_relative_same_file(fulltext_builder):
    schema = get_test_schema(
        a={
            "type": "object",
            OAREPO_MARSHMALLOW: {
                "schema-class": ".B",
                "generate": True,
            },
            "ui": {
                OAREPO_MARSHMALLOW: {
                    "schema-class": ".BUISchema",
                    "generate": True,
                }
            },
            "properties": {
                "b": {
                    "type": "keyword",
                }
            },
        }
    )
    fulltext_builder.filesystem = InMemoryFileSystem()
    fulltext_builder.build(schema, output_dir="")

    with fulltext_builder.filesystem.open(
        os.path.join("test", "services", "records", "ui_schema.py")
    ) as f:
        data = f.read()

    assert (
        re.sub(
            r"\s",
            "",
            """class TestUISchema(ma.Schema):
        \"""TestUISchema schema.\"""

        a = ma_fields.Nested(lambda:BUISchema())""",
        )
        in re.sub(r"\s", "", data)
    )


def test_generate_nested_schema_relative_upper(fulltext_builder):
    schema = get_test_schema(
        a={
            "type": "object",
            OAREPO_MARSHMALLOW: {
                "schema-class": "..schema2.B",
                "generate": True,
            },
            "ui": {
                OAREPO_MARSHMALLOW: {
                    "schema-class": "..schema2.BUISchema",
                    "generate": True,
                }
            },
            "properties": {
                "b": {
                    "type": "keyword",
                }
            },
        }
    )
    fulltext_builder.filesystem = InMemoryFileSystem()
    fulltext_builder.build(schema, output_dir="")

    with fulltext_builder.filesystem.open(
        os.path.join("test", "services", "records", "ui_schema.py")
    ) as f:
        data = f.read()

    assert (
        re.sub(
            r"\s",
            "",
            """class TestUISchema(ma.Schema):
        \"""TestUISchema schema.\"""

        a = ma_fields.Nested(lambda: BUISchema())""",
        )
        in re.sub(r"\s", "", data)
    )

    assert "from test.services.schema2 import BUISchema" in data


def test_generate_json_serializer(fulltext_builder):
    schema = get_test_schema(
        a={
            "type": "keyword",
        }
    )
    fulltext_builder.filesystem = InMemoryFileSystem()
    fulltext_builder.build(schema, output_dir="")
    with fulltext_builder.filesystem.open(
        os.path.join("test", "resources", "records", "ui.py")
    ) as f:
        data = f.read()
        print(data)
    assert (
        re.sub(
            r"\s",
            "",
            """
from flask_resources import BaseListSchema, MarshmallowSerializer
from flask_resources.serializers import JSONSerializer

from test.services.records.ui_schema import TestUISchema



class TestUIJSONSerializer(MarshmallowSerializer):
    \"""UI JSON serializer.\"""

    def __init__(self):
        \"""Initialise Serializer.\"""
        super().__init__(
            format_serializer_cls=JSONSerializer,
            object_schema_cls=TestUISchema,
            list_schema_cls=BaseListSchema,
            schema_context={"object_key": "ui"},
        )    
    """,
        )
        == re.sub(r"\s", "", data)
    )


def test_generate_resource_config(fulltext_builder):
    schema = get_test_schema(
        a={
            "type": "keyword",
        }
    )
    fulltext_builder.filesystem = InMemoryFileSystem()
    fulltext_builder.build(schema, output_dir="")
    with fulltext_builder.filesystem.open(
        os.path.join("test", "resources", "records", "config.py")
    ) as f:
        data = f.read()
        print(data)
    assert (
        re.sub(
            r"\s",
            "",
            """
import importlib_metadata
from flask_resources import ResponseHandler

from test.resources.records.ui import TestUIJSONSerializer



class TestResourceConfig():
    \"""TestRecord resource config.\"""

    blueprint_name = 'Test'
    url_prefix = '/test/'

    @property
    def response_handlers(self):
        entrypoint_response_handlers = {}
        for x in importlib_metadata.entry_points(group='invenio.test.response_handlers'):
            entrypoint_response_handlers.update(x.load())
        return {
            "application/vnd.inveniordm.v1+json": ResponseHandler(TestUIJSONSerializer()),
            **super().response_handlers,
            **entrypoint_response_handlers
        }    """,
        )
        == re.sub(r"\s", "", data)
    )


def test_localized_date(fulltext_builder):
    schema = get_test_schema(a={"type": "date"})
    fulltext_builder.filesystem = InMemoryFileSystem()
    fulltext_builder.build(schema, output_dir="")

    with fulltext_builder.filesystem.open(
        os.path.join("test", "services", "records", "ui_schema.py")
    ) as f:
        data = f.read()
    assert "a = l10n.LocalizedDate()" in data


def test_metadata(fulltext_builder):
    schema = get_test_schema(
        metadata={"type": "object", "properties": {"a": {"type": "keyword"}}}
    )
    fulltext_builder.filesystem = InMemoryFileSystem()
    fulltext_builder.build(schema, output_dir="")

    with fulltext_builder.filesystem.open(
        os.path.join("test", "services", "records", "ui_schema.py")
    ) as f:
        data = f.read()
    assert "metadata = ma_fields.Nested(lambda: TestMetadataUISchema())" in data



from .python_name import convert_name_to_python


def facet_definition(obj):
    key = None
    field = None
    args = None
    path = None
    if "facets" in obj.definition:
        facets = obj.definition["facets"]
        key = facets.get("key", obj.key)
        field = facets.get("field", None)
        args = facets.get("args", None)
        path = facets.get("path", None)
    if not key:
        key = obj.key
    return key, field, args, path


def facet_name(path):
    path.replace(".", "_")
    path = convert_name_to_python(path)
    return path
